#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from re import compile as re_compile
from typing import Any, Callable, Iterable, List, Optional, Pattern, Tuple, Union

from ._statics import TraceNames
from ._utils import TraceCallerData
from .api import API
from .async_api import AsyncAPI
from .logger import Logger
from .model import BotCommandObject, Model
from .session import AbstractSessionManager, SessionPatcher


class Plugins:
    _commands: List[BotCommandObject] = []
    _preprocessors: List[Callable[[Model.MESSAGE], Any]] = []
    api: Union[API, AsyncAPI] = TraceCallerData(TraceNames, ("self", "api"))
    logger: Logger = TraceCallerData(TraceNames, ("self", "logger"))
    session: AbstractSessionManager = SessionPatcher()

    def __new__(
        cls,
        bot_id=None,
        api=None,
        logger=None,
    ) -> Tuple[List, List]:
        commands, preprocessors = cls._commands, cls._preprocessors
        cls._commands, cls._preprocessors = [], []
        return commands, preprocessors

    @classmethod
    def get_preprocessor_names(cls):
        return (func.__name__ for func in cls._preprocessors)

    @classmethod
    def get_commands_names(cls):
        return (x.func.__name__ for x in cls._commands)

    @classmethod
    def before_command(cls):
        """
        注册plugins预处理器，将在检查所有commands前执行
        """

        def wrap(func: Callable[[Model.MESSAGE], Any]):
            cls._preprocessors.append(func)
            return func

        return wrap

    @classmethod
    def on_command(
        cls,
        command: Optional[Union[Iterable[str], str]] = None,
        regex: Optional[Union[Pattern, str, Iterable[Union[Pattern, str]]]] = None,
        is_treat: bool = True,
        is_require_at: bool = False,
        is_short_circuit: bool = True,
        is_custom_short_circuit: bool = False,
        is_require_admin: bool = False,
        admin_error_msg: Optional[str] = None,
    ):
        """
        注册plugins指令装饰器，可用于分割式编写指令并注册进机器人

        :param command: 可触发事件的指令列表，与正则regex互斥，优先使用此项
        :param regex: 可触发指令的正则compile实例或正则表达式，与指令表互斥
        :param is_treat: 是否在treated_msg中同时处理指令，如正则将返回.groups()，默认是
        :param is_require_at: 是否要求必须艾特机器人才能触发指令，默认否
        :param is_short_circuit: 如果触发指令成功是否短路不运行后续指令（将根据注册顺序排序指令的短路机制），默认是
        :param is_custom_short_circuit: 如果触发指令成功而回调函数返回True则不运行后续指令，存在时优先于is_short_circuit，默认否
        :param is_require_admin: 是否要求频道主或或管理才可触发指令，默认否
        :param admin_error_msg: 当is_require_admin为True，而触发用户的权限不足时，如此项不为None，返回此消息并短路；否则不进行短路
        """

        def wrap(func: Callable[[Model.MESSAGE], Any]):
            if is_short_circuit and is_custom_short_circuit:
                print(
                    "注意is_short_circuit与is_custom_short_circuit同时存在，将优先使用is_custom_short_circuit"
                )
            _kwargs = {
                "func": func,
                "treat": is_treat,
                "at": is_require_at,
                "short_circuit": is_short_circuit,
                "is_custom_short_circuit": is_custom_short_circuit,
                "admin": is_require_admin,
                "admin_error_msg": admin_error_msg,
            }
            if command:
                if isinstance(command, str):
                    command_obj = BotCommandObject(command=[command], **_kwargs)
                elif isinstance(command, Iterable):
                    command_obj = BotCommandObject(command=command, **_kwargs)
                else:
                    raise TypeError("command参数仅接受str或list类型的指令内容")
            else:
                if isinstance(regex, str):
                    command_obj = BotCommandObject(regex=[re_compile(regex)], **_kwargs)
                elif isinstance(regex, Pattern):
                    command_obj = BotCommandObject(regex=[regex], **_kwargs)
                elif isinstance(regex, Iterable):
                    regexs = []
                    for x in regex:
                        if isinstance(x, str):
                            regexs.append(re_compile(x))
                        elif isinstance(x, Pattern):
                            regexs.append(x)
                        else:
                            raise TypeError("regex参数仅接受re.compile返回的实例或str类型的正则表达式")
                    command_obj = BotCommandObject(regex=regexs, **_kwargs)
                else:
                    raise TypeError("regex参数仅接受re.compile返回的实例或str类型的正则表达式")
            cls._commands.append(command_obj)
            return func

        return wrap



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import asyncio
import sys
from unittest import mock

import pytest

import qg_botsdk

from ._base_context import (
    _async_callback,
    _callback,
    bot,
    bot_async,
    config,
    plugin_path,
)

pytest_plugins = ("pytest_asyncio",)

MockMsg = (
    '{"op":0,"s":2,"t":"MESSAGE_CREATE","id":"MESSAGE_CREATE:xxx",'
    '"d":{"author":{"avatar":"xxx","bot":false,"id":"xxx","username":"xxx"},'
    '"channel_id":"channel_id","content":"plugins_test","guild_id":"xxx","id":"id",'
    '"member":{"joined_at":"xxx","nick":"xxx","roles":[]},'
    '"mentions":[],"seq":9555,"seq_in_channel":"9555","timestamp":"xxx"}}'
)

MockMsgAdmin = (
    '{"op":0,"s":2,"t":"MESSAGE_CREATE","id":"MESSAGE_CREATE:xxx",'
    '"d":{"author":{"avatar":"xxx","bot":false,"id":"xxx","username":"xxx"},'
    '"channel_id":"channel_id","content":"plugins_test","guild_id":"xxx","id":"id",'
    '"member":{"joined_at":"xxx","nick":"xxx","roles":["2"]},'
    '"mentions":[],"seq":9555,"seq_in_channel":"9555","timestamp":"xxx"}}'
)

MockOp10Msg = '{"op": 10,"d": {"heartbeat_interval": 45000}}'


@pytest.mark.run_order(2)
class TestRunning:
    bot: qg_botsdk.BOT
    _called_counter: int = 0
    _last_called: float = 0

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_close(self, bot):
        self.bot = bot
        assert not self.bot.running
        with mock.patch.object(self.bot.logger, "error") as mock_logger:
            self.bot.close()
            mock_logger.assert_called_once()
            assert mock_logger.call_args == mock.call("当前机器人没有运行！")

    async def _start_assertions_with_proper_close(self):
        await asyncio.sleep(1)

        assert self.bot.running
        assert self.bot._bot_class is not None
        assert self.bot._bot_class.running
        assert self.bot._bot_class.session_id
        assert self.bot._bot_class.heartbeat
        assert (
            self.bot._bot_class.auth == f'Bot {config["bot_id"]}.{config["bot_token"]}'
        )
        assert self.bot.robot == self.bot._bot_class.robot
        self.bot.close()

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_start(self, bot):
        self.bot = bot
        assert not self.bot.running
        self.bot.start(is_blocking=False)
        self.bot.loop.run_until_complete(asyncio.sleep(0.5))
        self.bot.loop.create_task(self._start_assertions_with_proper_close())
        self.bot.block()

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_async_start(self, bot_async):
        self.bot = bot_async
        assert not self.bot.running
        self.bot.start(is_blocking=False)
        self.bot.loop.run_until_complete(asyncio.sleep(0.5))
        self.bot.loop.create_task(self._start_assertions_with_proper_close())
        self.bot.block()

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_start_more_than_once(self, bot):
        self.bot = bot

        with mock.patch.object(self.bot.logger, "error") as mock_logger:
            self.bot.block()
            mock_logger.assert_called_once()
            assert mock_logger.call_args == mock.call("当前机器人没有运行！")

        self.bot.start(is_blocking=False)
        self.bot.loop.run_until_complete(asyncio.sleep(0.5))

        with mock.patch.object(self.bot.logger, "error") as mock_logger:
            self.bot.start()
            mock_logger.assert_called_once()
            assert mock_logger.call_args == mock.call("当前机器人已在运行中！")

        self.bot.close()

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_reconnect(self, bot):
        self.bot = bot
        self.bot.start(is_blocking=False)
        self.bot.loop.run_until_complete(asyncio.sleep(0.5))
        while self.bot._bot_class.heartbeat not in asyncio.all_tasks(self.bot.loop):
            self.bot.loop.run_until_complete(asyncio.sleep(0.1))
        with mock.patch.object(self.bot._bot_class, "ws_send") as mock_ws_send:
            self.bot._bot_class.disable_reconnect = True
            self.bot.loop.create_task(self.bot._bot_class.dispatch_events(MockOp10Msg))
            self.bot.loop.run_until_complete(asyncio.sleep(1))
            mock_ws_send.assert_called_once()
            if sys.version_info >= (3, 8):
                assert '"op": 2' in mock_ws_send.call_args.args[0]
            self.bot._bot_class.disable_reconnect = False
            self.bot._bot_class.is_reconnect = True
            self.bot.loop.create_task(self.bot._bot_class.dispatch_events(MockOp10Msg))
            self.bot.loop.run_until_complete(asyncio.sleep(1))
            assert mock_ws_send.call_count == 2
            if sys.version_info >= (3, 8):
                assert '"op": 6' in mock_ws_send.call_args.args[0]

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_exception_on_ws(self, bot):
        self.bot = bot
        self.bot.start(is_blocking=False)
        self.bot.loop.run_until_complete(asyncio.sleep(0.5))
        while self.bot._bot_class.heartbeat not in asyncio.all_tasks(self.bot.loop):
            self.bot.loop.run_until_complete(asyncio.sleep(0.1))
        with mock.patch.object(self.bot.logger, "error") as mock_logger:
            with mock.patch.object(self.bot._bot_class.ws, "receive") as mock_ws_recv:
                mock_ws_recv.side_effect = ValueError("testing error")
                self.bot._bot_class.disable_reconnect = False
                self.bot.loop.create_task(
                    self.bot._bot_class.dispatch_events(MockOp10Msg)
                )
                while mock_logger.call_count < 2:
                    self.bot.loop.run_until_complete(asyncio.sleep(0.1))
                assert (
                    mock.call(repr(ValueError("testing error")))
                    in mock_logger.call_args_list
                )

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_bind_callback_and_plugins(self, bot):
        self.bot = bot
        self.bot._retrieve_new_plugins()
        self.bot.clear_current_plugins()
        self.bot.bind_msg(_callback)
        self.bot.load_plugins(plugin_path)
        self.bot.load_default_msg_logger()
        self.bot.start(is_blocking=False)
        self.bot.loop.run_until_complete(asyncio.sleep(0.5))
        while self.bot._bot_class.heartbeat not in asyncio.all_tasks(self.bot.loop):
            self.bot.loop.run_until_complete(asyncio.sleep(0.1))
        assert len(self.bot.get_current_preprocessors) == 2
        with mock.patch.object(self.bot.logger, "info") as mock_logger:
            with mock.patch.object(self.bot.api, "send_msg") as mock_send_msg:
                mock_send_msg.return_value = None
                self.bot.loop.create_task(self.bot._bot_class.dispatch_events(MockMsg))
                while not mock_send_msg.called:
                    self.bot.loop.run_until_complete(asyncio.sleep(0.1))
                self.bot.loop.run_until_complete(asyncio.sleep(0.5))
                assert (
                    mock.call("before_command_test", "plugins_test")
                    in mock_logger.call_args_list
                    and mock.call("收到频道 xxx 用户 xxx(xxx) 的消息：plugins_test")
                    in mock_logger.call_args_list
                )
                call_hist = mock_send_msg.call_args_list
                assert len(call_hist) == 2
                assert (
                    mock.call(
                        content="plugins_test",
                        image=None,
                        file_image=None,
                        message_reference_id=None,
                        ignore_message_reference_error=None,
                        message_id="id",
                        channel_id="channel_id",
                    )
                    in call_hist
                )
                assert (
                    mock.call(
                        content="test",
                        image=None,
                        file_image=None,
                        message_reference_id=None,
                        ignore_message_reference_error=None,
                        message_id="id",
                        channel_id="channel_id",
                    )
                    in call_hist
                )

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_async_bind_callback_and_plugins(self, bot_async):
        self.bot = bot_async
        self.bot._retrieve_new_plugins()
        self.bot.clear_current_plugins()
        self.bot.bind_msg(_async_callback)
        self.bot.start(is_blocking=False)
        self.bot.loop.run_until_complete(asyncio.sleep(0.5))
        while self.bot._bot_class.heartbeat not in asyncio.all_tasks(self.bot.loop):
            self.bot.loop.run_until_complete(asyncio.sleep(0.1))
        with mock.patch.object(self.bot.api, "send_msg") as mock_send_msg:
            mock_send_msg.return_value = None
            self.bot.loop.create_task(self.bot._bot_class.dispatch_events(MockMsg))
            while not mock_send_msg.called:
                self.bot.loop.run_until_complete(asyncio.sleep(0.1))
            self.bot.loop.run_until_complete(asyncio.sleep(0.5))
            mock_send_msg.assert_called_once()
            assert mock_send_msg.call_args == mock.call(
                content="test",
                image=None,
                file_image=None,
                message_reference_id=None,
                ignore_message_reference_error=None,
                message_id="id",
                channel_id="channel_id",
            )

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_regex_with_admin(self, bot):
        self.bot = bot
        self.bot._retrieve_new_plugins()
        self.bot.clear_current_plugins()
        self.bot.on_command(
            regex="(.+?)test", is_require_admin=True, admin_error_msg="err"
        )(_callback)
        self.bot.start(is_blocking=False)
        bot.loop.run_until_complete(asyncio.sleep(0.5))
        while self.bot._bot_class.heartbeat not in asyncio.all_tasks(self.bot.loop):
            self.bot.loop.run_until_complete(asyncio.sleep(0.1))
        with mock.patch.object(self.bot.api, "send_msg") as mock_send_msg:
            mock_send_msg.return_value = None
            for i, (mock_msg, ret_msg) in enumerate(
                ((MockMsg, "err"), (MockMsgAdmin, "test"))
            ):
                self.bot.loop.create_task(self.bot._bot_class.dispatch_events(mock_msg))
                while mock_send_msg.call_count <= i:
                    self.bot.loop.run_until_complete(asyncio.sleep(0.1))
                assert mock_send_msg.call_args == mock.call(
                    content=ret_msg,
                    image=None,
                    file_image=None,
                    message_reference_id=None,
                    ignore_message_reference_error=None,
                    message_id="id",
                    channel_id="channel_id",
                )

    def _callback_with_create_dynamic_command(self, data: qg_botsdk.Model.MESSAGE):
        data.reply("dynamic_command_test")
        self.bot.on_command(
            regex="(.+?)test", is_require_admin=True, admin_error_msg="err"
        )(_callback)
        return True

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_dynamic_created_command_on_runtime(self, bot):
        self.bot = bot
        self.bot._retrieve_new_plugins()
        self.bot.clear_current_plugins()
        self.bot.on_command("plugins_test", is_custom_short_circuit=True)(
            self._callback_with_create_dynamic_command
        )
        self.bot.start(is_blocking=False)
        bot.loop.run_until_complete(asyncio.sleep(0.5))
        while self.bot._bot_class.heartbeat not in asyncio.all_tasks(self.bot.loop):
            self.bot.loop.run_until_complete(asyncio.sleep(0.1))
        with mock.patch.object(self.bot.api, "send_msg") as mock_send_msg:
            mock_send_msg.return_value = None
            self.bot.loop.create_task(self.bot._bot_class.dispatch_events(MockMsg))
            while not mock_send_msg.called:
                self.bot.loop.run_until_complete(asyncio.sleep(0.1))
            self.bot.loop.run_until_complete(asyncio.sleep(0.5))
            assert mock_send_msg.call_count == 1
            assert mock_send_msg.call_args == mock.call(
                content="dynamic_command_test",
                image=None,
                file_image=None,
                message_reference_id=None,
                ignore_message_reference_error=None,
                message_id="id",
                channel_id="channel_id",
            )
            assert len(self.bot.get_current_commands) == 2

    async def _test_wait_for_msg_binder(self, data):
        command_obj = qg_botsdk.BotCommandObject(["plugins_test"], treat=False)
        _data = await self.bot.api.wait_for(
            qg_botsdk.Scope.USER, command_obj, timeout=3
        )
        assert _data.content == "plugins_test"
        self._called_counter = 1

    async def _test_wait_for_content(self):
        self._called_counter = 0
        for i in range(2):
            self.bot.loop.create_task(self.bot._bot_class.dispatch_events(MockMsg))
            while not self._called_counter < 1:
                self.bot.loop.run_until_complete(asyncio.sleep(0.1))
            self._called_counter = 0

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_wait_for(self, bot_async):
        self.bot = bot_async
        self.bot.bind_msg(self._test_wait_for_msg_binder)
        self.bot.start(is_blocking=False)
        self.bot.loop.run_until_complete(asyncio.sleep(0.5))
        while self.bot._bot_class.heartbeat not in asyncio.all_tasks(self.bot.loop):
            self.bot.loop.run_until_complete(asyncio.sleep(0.1))
        self.bot.loop.run_until_complete(self._test_wait_for_content())

    def _loop_event(self):
        self.bot.api.send_msg("channel_id", "content test")
        self._called_counter += 1
        time_now = self.bot.loop.time()
        assert time_now - self._last_called >= 1
        self._last_called = time_now

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_loop_event(self, bot):
        self.bot = bot
        bot.register_repeat_event(self._loop_event, 1)
        bot.start(is_blocking=False)
        bot.loop.run_until_complete(asyncio.sleep(0.5))
        while self.bot._bot_class.heartbeat not in asyncio.all_tasks(self.bot.loop):
            self.bot.loop.run_until_complete(asyncio.sleep(0.1))
        with mock.patch.object(self.bot.api, "send_msg") as mock_send_msg:
            mock_send_msg.return_value = None
            while self._called_counter < 2:
                self.bot.loop.run_until_complete(asyncio.sleep(0.1))
            self._called_counter = 0
            assert mock_send_msg.call_count == 2
            assert mock_send_msg.call_args_list[0] == mock.call(
                "channel_id", "content test"
            )
            assert mock_send_msg.call_args_list[1] == mock.call(
                "channel_id", "content test"
            )

    def _start_event(self):
        self._called_counter += 1
        assert self.bot.api.get_bot_info().result

    @pytest.mark.asyncio
    @pytest.mark.timeout(20)
    def test_start_event(self, bot):
        self.bot = bot
        bot.register_start_event(self._start_event)
        bot.start(is_blocking=False)
        bot.loop.run_until_complete(asyncio.sleep(5))
        while self.bot._bot_class.heartbeat not in asyncio.all_tasks(self.bot.loop):
            self.bot.loop.run_until_complete(asyncio.sleep(0.1))
        assert self._called_counter == 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from qg_botsdk import Model, Plugins


@Plugins.before_command()
def before_command_test(data: Model.MESSAGE):
    Plugins.logger.info("before_command_test", data.treated_msg)


@Plugins.on_command("plugins_test", is_short_circuit=False)
def plugins_test(data: Model.MESSAGE):
    data.reply("plugins_test")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import pickle
from asyncio import AbstractEventLoop, sleep
from collections import namedtuple
from copy import deepcopy
from threading import Thread
from time import time
from typing import Dict, Hashable, Iterable, List, Optional, Union
from weakref import finalize

from ._statics import EVENTS
from ._utils import exception_handler
from .api import API
from .async_api import AsyncAPI
from .logger import Logger
from .model import (
    BotCommandObject,
    Model,
    Scope,
    SessionObject,
    SessionStatus,
    WaifForCommandCallback,
)

_AllScopeStr = ("USER", "GUILD", "CHANNEL", "GLOBAL")
ScopeRegisterKey = namedtuple("ScopeRegisterKey", _AllScopeStr)


class _SessionObject:
    def __init__(
        self,
        status: SessionStatus,
        data: Dict,
        timeout: Optional[float] = None,
        last_operate: float = 0,
        timeout_reply: str = None,
        inactive_gc_timeout: Optional[float] = None,
        gc_timeout_stamp: Optional[float] = None,
        timeout_reply_params: Dict = None,
        timeout_reply_message_id_expire: float = None,
    ):
        self.status = status
        self.data = data
        self.timeout = timeout
        self.last_operate = last_operate
        self.timeout_reply = timeout_reply
        self.inactive_gc_timeout = inactive_gc_timeout  # 处于INACTIVE的session在相隔这段时间后被回收
        self.gc_timeout_stamp = gc_timeout_stamp  # 处于INACTIVE的session在此时间后被回收
        self.timeout_reply_params = timeout_reply_params  # 超时回复的消息默认参数
        self.timeout_reply_message_id_expire = (
            timeout_reply_message_id_expire  # 超时回复的消息ID过期时间
        )

    def __repr__(self):
        return (
            f"<SessionObject status={self.status} data={self.data} timeout={self.timeout} "
            f"last_operate={self.last_operate} timeout_reply={self.timeout_reply} "
            f"inactive_gc_timeout={self.inactive_gc_timeout} gc_timeout_stamp={self.gc_timeout_stamp} "
            f"timeout_reply_params={self.timeout_reply_params} "
            f"timeout_reply_message_id_expire={self.timeout_reply_message_id_expire}>"
        )


class SessionManager:
    """
    Implementation of Session Manager, used to create and manage sessions with user, guild, channel and global scope.
    Life cycle of session: ACTIVE --(timeout || active end)--> INACTIVE --(inactive_gc_timeout)--> GC

    Structure of sessions:
        - normal sessions(e.g. user) : {"user_id": {"key": _SessionObject, ...}, ...}
        - global sessions struct: {"key": _SessionObject, ...}
    """

    def __init__(
        self,
        logger: Logger,
        commit_path: Optional[str] = None,
        is_auto_commit: bool = True,
    ):
        self.__sessions: Dict = {x: {} for x in _AllScopeStr}
        self.__sessions_pk_cache = None
        self.__wait_for_registers: Dict = {}
        self.__logger: Logger = logger
        self.__bot_identify: str = logger.bot_app_id
        self.__commit_path: str = commit_path or os.path.join(
            os.getcwd(), "session_data"
        )
        self.__check_path(self.__commit_path)
        self.__is_auto_commit = is_auto_commit
        self.__is_running = False
        self.fetch_data()
        self.api = None  # what a cursed implementation...
        self.__finalizer = finalize(
            self, lambda x: x(), self.__del__
        )  # ensure commit_data() is called before exit

    def __del__(self):
        if self.__is_auto_commit:
            try:
                self.__logger.info("Session Manager结束中，正在保存数据...")
            except Exception:
                print("Session Manager结束中，正在保存数据...")
            self.commit_data(is_info=False)

    # -*- class internal methods -*-
    @staticmethod
    def __check_path(commit_path: str):
        if not os.path.exists(commit_path):
            path_splits = os.path.split(commit_path)
            if "." in path_splits[-1]:
                raise ValueError(f"路径 '{commit_path}' 不是文件夹（包含'.'）")
            os.makedirs(commit_path)
        elif not os.path.isdir(commit_path):
            raise ValueError(f"路径 '{commit_path}' 不是文件夹")

    @staticmethod
    def __end_session(session: _SessionObject):
        session.status = SessionStatus.INACTIVE
        gc_timeout = session.inactive_gc_timeout
        session.gc_timeout_stamp = (
            time() + gc_timeout if gc_timeout is not None else None
        )

    @staticmethod
    def __update_last_op(session: _SessionObject):
        session.last_operate = time()
        if session.status != SessionStatus.ACTIVE:
            session.status = SessionStatus.ACTIVE

    def __manage_session_object(self, loop, time_now, session, scope, identify):
        _change = False
        gc_keys = []
        for k, v in session.items():
            if (
                v.status == SessionStatus.ACTIVE
                and v.timeout is not None
                and time_now - v.last_operate > v.timeout
            ):
                if v.timeout_reply:
                    try:
                        if time_now > v.timeout_reply_message_id_expire:
                            del v.timeout_reply_params["message_id"]
                        _params = {"content": v.timeout_reply, **v.timeout_reply_params}
                        if isinstance(self.api, AsyncAPI):
                            loop.create_task(self.api.send_msg(**_params))

                        elif isinstance(self.api, API):
                            Thread(
                                target=self.api.send_msg, kwargs=_params, daemon=True
                            ).start()
                        else:
                            pass
                    except Exception as e:
                        self.__logger.error(
                            f"Session({scope}::{k}) 超时回调错误：{e.__repr__()}"
                        )
                        self.__logger.error(exception_handler(e))
                if identify:
                    self.__logger.info(f"Session({scope}::{identify}::{k}) 已超时")
                else:
                    self.__logger.info(f"Session({scope}::{k}) 已超时")
                self.__end_session(v)
                _change = True
            if (
                v.status == SessionStatus.INACTIVE
                and v.gc_timeout_stamp is not None
                and time_now > v.gc_timeout_stamp
            ):
                gc_keys.append(k)
        for k in gc_keys:
            del session[k]
            _change = True
        return _change

    async def __manager_loop(self, loop: AbstractEventLoop):
        while True:
            await sleep(0.1)
            change = False
            try:
                time_now = time()
                for scope in _AllScopeStr:
                    sessions = self.__check_scope(scope)
                    if scope == "GLOBAL":
                        change = self.__manage_session_object(
                            loop, time_now, sessions, scope, None
                        )
                    else:
                        gc_identifies = []
                        for identify, session in sessions.items():
                            change = self.__manage_session_object(
                                loop, time_now, session, scope, identify
                            )
                            if not session:
                                gc_identifies.append(identify)
                        for identify in gc_identifies:
                            del sessions[identify]
                            change = True
            except Exception as e:
                self.__logger.error(e.__repr__())
                self.__logger.error(exception_handler(e))
            if self.__is_auto_commit:
                if change:
                    self.commit_data(is_info=False)
                else:
                    _new_pk_data = pickle.dumps(self.__sessions)
                    if self.__sessions_pk_cache != _new_pk_data:
                        self.__sessions_pk_cache = _new_pk_data
                        self.commit_data(is_info=False, pk_data=_new_pk_data)

    @staticmethod
    def __valid_scope(scope):
        if isinstance(scope, Scope):
            scope = scope.value
        elif scope in _AllScopeStr:
            pass
        else:
            raise ValueError("scope参数仅接受user, guild, channel, global")
        return scope

    def __check_identify(self, scope: Union[Scope, str], data) -> Hashable:
        if scope == Scope.USER or scope == "USER":
            identify = self.__get_author_id(data)
        elif scope == Scope.GUILD or scope == "GUILD":
            identify = self.__get_guild_id(data)
        elif scope == Scope.CHANNEL or scope == "CHANNEL":
            identify = self.__get_channel_id(data)
        else:
            identify = None
        return identify

    def __check_scope(
        self, scope: Union[Scope, str]
    ) -> Dict[Hashable, Dict[Hashable, _SessionObject]]:
        if scope == Scope.USER or scope == "USER":
            target_sessions = self.__sessions["USER"]
        elif scope == Scope.GUILD or scope == "GUILD":
            target_sessions = self.__sessions["GUILD"]
        elif scope == Scope.CHANNEL or scope == "CHANNEL":
            target_sessions = self.__sessions["CHANNEL"]
        else:
            target_sessions = self.__sessions["GLOBAL"]
        return target_sessions

    @staticmethod
    def __get_author_id(data) -> Hashable:
        t = data.t
        if t in EVENTS.MESSAGE_CREATE or t in EVENTS.DM_CREATE:
            return data.author.id
        elif t in EVENTS.GUILD or t in EVENTS.CHANNEL:
            return data.op_user_id
        elif t in EVENTS.GUILD_MEMBER:
            return data.user.id
        elif t in EVENTS.MESSAGE_DELETE:
            return data.op_user.id
        elif t in EVENTS.FORUM or t in EVENTS.OPEN_FORUM:
            return data.author_id
        elif t in EVENTS.REACTION or t in EVENTS.ALC_MEMBER:
            return data.user_id
        elif t in EVENTS.INTERACTION:
            return data.data.resolved.user_id
        else:
            raise ValueError(f"当前事件{data.t}不支持使用user scope")

    @staticmethod
    def __get_guild_id(data) -> Hashable:
        t = data.t
        if (
            t in EVENTS.MESSAGE_CREATE
            or t in EVENTS.DM_CREATE
            or t in EVENTS.CHANNEL
            or t in EVENTS.GUILD_MEMBER
            or t in EVENTS.AUDIT
            or t in EVENTS.FORUM
            or t in EVENTS.OPEN_FORUM
            or t in EVENTS.AUDIO
            or t in EVENTS.REACTION
            or t in EVENTS.INTERACTION
            or t in EVENTS.ALC_MEMBER
        ):
            return data.guild_id
        elif t in EVENTS.GUILD:
            return data.id
        elif t in EVENTS.MESSAGE_DELETE:
            return data.message.guild_id
        else:
            raise ValueError(f"当前事件{data.t}不支持使用guild scope")

    @staticmethod
    def __get_channel_id(data) -> Hashable:
        t = data.t
        if (
            t in EVENTS.MESSAGE_CREATE
            or t in EVENTS.DM_CREATE
            or t in EVENTS.FORUM
            or t in EVENTS.OPEN_FORUM
            or t in EVENTS.AUDIO
            or t in EVENTS.REACTION
            or t in EVENTS.INTERACTION
            or t in EVENTS.ALC_MEMBER
        ):
            return data.channel_id
        elif t in EVENTS.CHANNEL:
            return data.id
        elif t in EVENTS.MESSAGE_DELETE:
            return data.message.channel_id
        else:
            raise ValueError(f"当前事件{data.t}不支持使用channel scope")

    def __check_and_get_target_session(self, obj, scope, key, identify):
        if not identify:
            identify = self.__check_identify(scope, obj)
        scope = self.__valid_scope(scope)
        target_sessions = self.__check_scope(scope)
        if identify:
            if identify not in target_sessions or key not in target_sessions[identify]:
                raise KeyError(f"Scope {scope} 中不存在 {identify}::{key} 的session")
            target_session = target_sessions[identify][key]
        else:
            if key not in target_sessions:
                raise KeyError(f"Scope {scope} 中不存在 {key} 的session")
            target_session = target_sessions[key]
        return target_session

    # -*- user methods -*-
    def new(
        self,
        obj,
        scope: Scope,
        key: Hashable,
        data: Dict = None,
        identify: Hashable = None,
        is_replace: bool = True,
        timeout: Optional[float] = None,
        timeout_reply: str = None,
        inactive_gc_timeout: Optional[float] = 0,
    ) -> SessionObject:
        if not identify:
            identify = self.__check_identify(scope, obj)
        scope = self.__valid_scope(scope)
        target_sessions = self.__check_scope(scope)
        if identify:
            if identify not in target_sessions:
                target_sessions[identify] = {}
            target_sessions = target_sessions[identify]
        if key in target_sessions and not is_replace:
            if identify:
                raise KeyError(f"Scope {scope} 中已存在 {identify}::{key} 的session")
            else:
                raise KeyError(f"Scope {scope} 中已存在 {key} 的session")
        target_sessions[key] = _SessionObject(
            status=SessionStatus.ACTIVE,
            data=data if data is not None else {},
            timeout=timeout,
            last_operate=time(),
            timeout_reply=timeout_reply,
            inactive_gc_timeout=inactive_gc_timeout,
            timeout_reply_params={"message_id": obj.id, "channel_id": obj.channel_id},
            timeout_reply_message_id_expire=time() + 300,
        )
        if self.__is_auto_commit:
            self.commit_data(is_info=False)
        return SessionObject(scope, SessionStatus.ACTIVE, key, data, identify)

    def get(
        self,
        obj,
        scope: Scope,
        key: Hashable,
        identify: Hashable = None,
        default=None,
    ) -> SessionObject:
        if not identify:
            identify = self.__check_identify(scope, obj)
        scope = self.__valid_scope(scope)
        target_sessions = self.__check_scope(scope)
        if identify:
            if identify not in target_sessions or key not in target_sessions[identify]:
                return default
            target_session = target_sessions[identify][key]
        else:
            if key not in target_sessions:
                return default
            target_session = target_sessions[key]
        self.__update_last_op(target_session)
        return SessionObject(
            scope, target_session.status, key, target_session.data, identify
        )

    def update(
        self,
        obj,
        scope: Scope,
        key: Hashable,
        data: Dict,
        identify: Hashable = None,
    ) -> SessionObject:
        target_session = self.__check_and_get_target_session(obj, scope, key, identify)
        target_session.data.update(data)
        self.__update_last_op(target_session)
        if self.__is_auto_commit:
            self.commit_data(is_info=False)
        return SessionObject(
            scope, target_session.status, key, target_session.data, identify
        )

    def remove(
        self, scope: Scope = None, identify: Hashable = None, key: Hashable = None
    ):
        if not scope:
            self.__sessions = {x: {} for x in _AllScopeStr}
            return
        scope = self.__valid_scope(scope)
        target_sessions = self.__check_scope(scope)
        if not key and not identify:
            target_sessions.clear()
            if self.__is_auto_commit:
                self.commit_data(is_info=False)
            return
        if not key and identify:
            target_sessions[identify] = {}
            if self.__is_auto_commit:
                self.commit_data(is_info=False)
            return
        if identify:
            if identify not in target_sessions or key not in target_sessions[identify]:
                raise KeyError(f"Scope {scope} 中不存在 {identify}::{key} 的session")
            target_sessions = target_sessions[identify]
        else:
            if key not in target_sessions:
                raise KeyError(f"Scope {scope} 中不存在 {key} 的session")
        target_sessions.pop(key)
        if self.__is_auto_commit:
            self.commit_data(is_info=False)

    def end(
        self,
        obj,
        scope: Scope,
        key: Hashable,
        identify: Hashable = None,
        inactive_gc_timeout: Optional[float] = 0,
    ) -> SessionObject:
        target_session = self.__check_and_get_target_session(obj, scope, key, identify)
        target_session.inactive_gc_timeout = inactive_gc_timeout
        self.__end_session(target_session)
        if self.__is_auto_commit:
            self.commit_data(is_info=False)
        return SessionObject(
            scope, target_session.status, key, target_session.data, identify
        )

    def get_all(self) -> Dict:
        return deepcopy(self.__sessions)

    def set_status(
        self,
        obj,
        scope: Scope,
        key: Hashable,
        status: SessionStatus,
        identify: Hashable = None,
    ):
        if not isinstance(status, SessionStatus):
            raise ValueError("status 必须是 SessionStatus 的枚举值")
        if status == SessionStatus.INACTIVE:
            self.end(obj, scope, key, identify)
            return
        target_session = self.__check_and_get_target_session(obj, scope, key, identify)
        target_session.status = status
        target_session.last_operate = time()
        if self.__is_auto_commit:
            self.commit_data(is_info=False)

    def get_status(
        self, obj, scope: Scope, key: Hashable, identify: Hashable = None
    ) -> SessionStatus:
        target_session = self.__check_and_get_target_session(obj, scope, key, identify)
        return target_session.status

    # -*- fetch/commit data methods -*-
    def fetch_data(self, is_info: bool = True):
        _path = os.path.join(self.__commit_path, f"session_{self.__bot_identify}.db")
        if not os.path.exists(_path):
            return
        try:
            with open(_path, "rb") as f_db:
                db = pickle.loads(f_db.read())
                for scope in _AllScopeStr:
                    try:
                        scope_data = db[scope]
                    except KeyError:
                        continue
                    except Exception as e:
                        self.__logger.error(
                            f"Session Manager 读取 Scope::{scope} 数据时出现错误，将跳过：{repr(e)}"
                        )
                    if scope == "GLOBAL":
                        for k in scope_data:
                            try:
                                self.__sessions[scope][k] = scope_data[k]
                            except Exception as e:
                                self.__logger.error(
                                    f"Session Manager 读取 Scope::{scope}::{k} 数据时出现错误，将跳过：{repr(e)}"
                                )
                    else:
                        for _id in scope_data:
                            if _id not in self.__sessions[scope]:
                                self.__sessions[scope][_id] = {}
                            for k in scope_data[_id]:
                                try:
                                    self.__sessions[scope][_id][k] = scope_data[_id][k]
                                except Exception as e:
                                    self.__logger.error(
                                        f"Session Manager 读取 Scope::{scope}::{_id}::{k} 数据时出现错误，将跳过：{repr(e)}"
                                    )
            if is_info:
                self.__logger.info(f"Session Manager 读取了 {self.__commit_path} 的数据")
        except Exception as e:
            self.__logger.error(f"Session Manager 读取 {_path} 时出现错误，将跳过：{repr(e)}")

    def commit_data(self, is_info: bool = True, pk_data: Optional[bytes] = None):
        _path = os.path.join(self.__commit_path, f"session_{self.__bot_identify}.db")
        try:
            with open(_path, "wb") as f_db:
                if not pk_data:
                    pk_data = pickle.dumps(self.__sessions)
                f_db.write(pk_data)
            if is_info:
                self.__logger.info(f"Session Manager 写入了 {_path} 的数据")
        except Exception as e:
            self.__logger.error(f"Session Manager 写入 {_path} 时出现错误，将跳过：{repr(e)}")

    def set_auto_commit(self, is_auto_commit: bool):
        self.__is_auto_commit = is_auto_commit

    def set_commit_path(self, commit_path: str):
        self.__check_path(commit_path)
        self.__commit_path = commit_path

    # -*- helper private methods for internal use -*-
    def start(self, loop: AbstractEventLoop):
        if not self.__is_running:
            loop.create_task(self.__manager_loop(loop))
            self.__is_running = True

    # -*- helper private methods for api.wait_for -*-
    def register_wait_for(
        self, obj, scopes: Union[Scope, Iterable[Scope]], command: BotCommandObject
    ):
        _scope_value = {x: None for x in _AllScopeStr}
        if isinstance(scopes, Iterable):
            for scope in scopes:
                if not isinstance(scope, Scope):
                    raise ValueError("scope 必须是 Scope 的枚举值")
                _scope_value[scope.value] = self.__check_identify(scope, obj)
        else:
            if not isinstance(scopes, Scope):
                raise ValueError("scope 必须是 Scope 的枚举值")
            _scope_value[scopes.value] = self.__check_identify(scopes, obj)
        _scope_key = ScopeRegisterKey(**_scope_value)
        if _scope_key not in self.__wait_for_registers:
            self.__wait_for_registers[_scope_key] = {}
        self.__wait_for_registers[_scope_key][command] = None
        return _scope_key

    def check_wait_for(self, scope_key: ScopeRegisterKey, command: BotCommandObject):
        if (
            scope_key not in self.__wait_for_registers
            or command not in self.__wait_for_registers[scope_key]
        ):
            return False, None
        data = self.__wait_for_registers[scope_key][command]
        if data:
            del self.__wait_for_registers[scope_key][command]
            if not self.__wait_for_registers[scope_key]:
                del self.__wait_for_registers[scope_key]
        return True, data

    def del_wait_for(self, scope_key: ScopeRegisterKey, command: BotCommandObject):
        if scope_key not in self.__wait_for_registers:
            return
        if command in self.__wait_for_registers[scope_key]:
            del self.__wait_for_registers[scope_key][command]
        if not self.__wait_for_registers[scope_key]:
            del self.__wait_for_registers[scope_key]

    def wait_for_message_checker(
        self, obj: Model.MESSAGE
    ) -> List[WaifForCommandCallback]:
        triggered_commands = []
        _scope_value = {
            scope: self.__check_identify(scope, obj) for scope in _AllScopeStr
        }
        for scope_key, command_callback in self.__wait_for_registers.items():
            flag = True
            for scope in _AllScopeStr:
                scope_key_value = getattr(scope_key, scope)
                if (
                    scope_key_value is not None
                    and scope_key_value != _scope_value[scope]
                ):
                    flag = False
                    break
            if flag:
                for command in command_callback:
                    triggered_commands.append(
                        WaifForCommandCallback(
                            command=command,
                            callback=lambda data: command_callback.__setitem__(
                                command, data
                            ),
                        )
                    )
        return triggered_commands



from io import BufferedReader
from os.path import exists
from typing import BinaryIO, Dict, List, Optional, Union

from ._api_model import BaseMessageApiModel
from ._api_model import MessageConstructRet as _MessageConstructRet
from ._utils import sdk_error_temp
from .http import FormData_


class ApiModel:
    class Message(BaseMessageApiModel):
        def __init__(
            self,
            content: Optional[str] = None,
            image: Optional[str] = None,
            file_image: Optional[Union[bytes, BinaryIO, str]] = None,
            message_reference_id: Optional[str] = None,
            ignore_message_reference_error: bool = False,
        ):
            """
            构建一个消息对象，用于发送消息

            :param content: 消息文本（选填，此项与image至少需要有一个字段，否则无法下发消息）
            :param image: 图片url，不可发送本地图片（选填，此项与msg至少需要有一个字段，否则无法下发消息）
            :param file_image: 本地图片，可选三种方式传参，具体可参阅github中的example_10或帮助文档，与image同时存在时优先使用此项
            :param message_reference_id: 引用消息的id（选填）
            :param ignore_message_reference_error: 是否忽略获取引用消息详情错误，默认否（选填）
            """
            super().__init__()
            if content is not None and not isinstance(content, str):
                content = str(content)
            self.__content = content
            self.__image = image
            self.__file_image = file_image
            self.__message_reference_id = message_reference_id
            if ignore_message_reference_error is None:
                ignore_message_reference_error = False
            self.__ignore_message_reference_error = ignore_message_reference_error

            self._constructed_obj = self.__construct(self._message_id, self._event_id)

        def __repr__(self):
            return (
                f"<Message content={self.__content}, image={self.__image}, file_image={self.__file_image}, "
                f"message_reference_id={self.__message_reference_id}, "
                f"ignore_message_reference_error={self.__ignore_message_reference_error}>"
            )

        def __construct(self, message_id, event_id) -> _MessageConstructRet:
            """
            internal construct method

            :return: kwargs for http api request
            """
            if self.__message_reference_id is not None:
                json_ = {
                    "content": self.__content,
                    "msg_id": message_id,
                    "event_id": event_id,
                    "image": self.__image,
                    "message_reference": {
                        "message_id": self.__message_reference_id,
                        "ignore_get_message_error": self.__ignore_message_reference_error,
                    },
                }
            else:
                json_ = {
                    "content": self.__content,
                    "msg_id": message_id,
                    "event_id": event_id,
                    "image": self.__image,
                }
            if self.__file_image is not None:
                if isinstance(self.__file_image, BufferedReader):
                    self.__file_image = self.__file_image.read()
                elif isinstance(self.__file_image, str):
                    if exists(self.__file_image):
                        with open(self.__file_image, "rb") as img:
                            self.__file_image = img.read()
                    else:
                        if self.__file_image.startswith("http"):
                            return _MessageConstructRet(
                                result=False,
                                error_ret=sdk_error_temp(
                                    "发送网络图片请使用image参数，而非file_image"
                                ),
                            )
                        return _MessageConstructRet(
                            result=False, error_ret=sdk_error_temp("目标图片路径不存在，无法发送")
                        )
                elif not isinstance(self.__file_image, bytes):
                    return _MessageConstructRet(
                        result=False,
                        error_ret=sdk_error_temp(
                            f"file_image不支持{type(self.__file_image)}的内容"
                        ),
                    )
                json_["file_image"] = self.__file_image
                data_ = FormData_()
                for keys, values in json_.items():
                    if values is not None and keys != "image":
                        data_.add_field(keys, values)
                return _MessageConstructRet(result=True, kwargs={"data": data_})
            else:
                return _MessageConstructRet(result=True, kwargs={"json": json_})

    class MessageEmbed(BaseMessageApiModel):
        def __init__(
            self,
            title: Optional[str] = None,
            content: Optional[List[str]] = None,
            image: Optional[str] = None,
            prompt: Optional[str] = None,
        ):
            """
            构建一个embed消息对象，用于发送消息

            :param title: 标题文本（选填）
            :param content: 内容文本列表，每一项之间将存在分行（选填）
            :param image: 略缩图url，不可发送本地图片（选填）
            :param prompt: 消息弹窗通知的文本内容（选填）
            """
            super().__init__()
            self.__title = title
            self.__content = content
            self.__image = image
            self.__prompt = prompt

            self._constructed_obj = self.__construct(self._message_id, self._event_id)

        def __repr__(self):
            return (
                f"<MessageEmbed title={self.__title}, content={self.__content}, image={self.__image}, "
                f"prompt={self.__prompt}>"
            )

        def __construct(self, message_id, event_id) -> _MessageConstructRet:
            """
            internal construct method

            :return: kwargs for http api request
            """
            json_ = {
                "embed": {
                    "title": self.__title,
                    "prompt": self.__prompt,
                    "thumbnail": {"url": self.__image},
                    "fields": [],
                },
                "msg_id": message_id,
                "event_id": event_id,
            }
            if self.__content is not None:
                for items in self.__content:
                    json_["embed"]["fields"].append({"name": str(items)})
            return _MessageConstructRet(result=True, kwargs={"json": json_})

    class MessageArk23(BaseMessageApiModel):
        def __init__(
            self,
            content: List[str],
            link: List[str],
            desc: Optional[str] = None,
            prompt: Optional[str] = None,
        ):
            """
            构建一个ark23消息对象，用于发送消息

            :param content: 内容文本列表，每一项之间将存在分行
            :param link: 链接url列表，长度应与内容列一致。将根据位置顺序填充文本超链接，如文本不希望填充链接可使用空文本或None填充位置
            :param desc: 描述文本内容（选填）
            :param prompt: 消息弹窗通知的文本内容（选填）
            """
            super().__init__()
            self.__content = content
            self.__link = link
            self.__desc = desc
            self.__prompt = prompt

            self._constructed_obj = self.__construct(self._message_id, self._event_id)

        def __repr__(self):
            return (
                f"<MessageArk23 content={self.__content}, link={self.__link}, desc={self.__desc}, "
                f"prompt={self.__prompt}>"
            )

        def __construct(self, message_id, event_id) -> _MessageConstructRet:
            """
            internal construct method

            :return: kwargs for http api request
            """
            if len(self.__content) != len(self.__link):
                return _MessageConstructRet(
                    result=False, error_ret=sdk_error_temp("注意内容列表长度应与链接列表长度一致")
                )
            json_ = {
                "ark": {
                    "template_id": 23,
                    "kv": [
                        {"key": "#DESC#", "value": self.__desc},
                        {"key": "#PROMPT#", "value": self.__prompt},
                        {"key": "#LIST#", "obj": []},
                    ],
                },
                "msg_id": message_id,
                "event_id": event_id,
            }
            for _link, _content in zip(self.__link, self.__content):
                if _content is not None and not isinstance(_content, str):
                    _content = str(_content)
                if _link is not None and not isinstance(_link, str):
                    _link = str(_link)
                json_["ark"]["kv"][2]["obj"].append(
                    {
                        "obj_kv": [
                            {"key": "desc", "value": _content},
                            {"key": "link", "value": _link},
                        ]
                    }
                )
            return _MessageConstructRet(result=True, kwargs={"json": json_})

    class MessageArk24(BaseMessageApiModel):
        def __init__(
            self,
            title: Optional[str] = None,
            content: Optional[str] = None,
            subtitile: Optional[str] = None,
            link: Optional[str] = None,
            image: Optional[str] = None,
            desc: Optional[str] = None,
            prompt: Optional[str] = None,
        ):
            """
            构建一个ark24消息对象，用于发送消息

            :param title: 标题文本（选填）
            :param content: 详情描述文本（选填）
            :param subtitile: 子标题文本（选填）
            :param link: 跳转的链接url（选填）
            :param image: 略缩图url，不可发送本地图片（选填）
            :param desc: 描述文本内容（选填）
            :param prompt: 消息弹窗通知的文本内容（选填）
            """
            super().__init__()
            self.__title = title
            self.__content = content
            self.__subtitile = subtitile
            self.__link = link
            self.__image = image
            self.__desc = desc
            self.__prompt = prompt

            self._constructed_obj = self.__construct(self._message_id, self._event_id)

        def __repr__(self):
            return (
                f"<MessageArk24 title={self.__title}, content={self.__content}, subtitile={self.__subtitile}, "
                f"link={self.__link}, image={self.__image}, desc={self.__desc}, prompt={self.__prompt}>"
            )

        def __construct(self, message_id, event_id) -> _MessageConstructRet:
            """
            internal construct method

            :return: kwargs for http api request
            """
            json_ = {
                "ark": {
                    "template_id": 24,
                    "kv": [
                        {"key": "#DESC#", "value": self.__desc},
                        {"key": "#PROMPT#", "value": self.__prompt},
                        {"key": "#TITLE#", "value": self.__title},
                        {"key": "#METADESC#", "value": self.__content},
                        {"key": "#IMG#", "value": self.__image},
                        {"key": "#LINK#", "value": self.__link},
                        {"key": "#SUBTITLE#", "value": self.__subtitile},
                    ],
                },
                "msg_id": message_id,
                "event_id": event_id,
            }
            return _MessageConstructRet(result=True, kwargs={"json": json_})

    class MessageArk37(BaseMessageApiModel):
        def __init__(
            self,
            title: Optional[str] = None,
            content: Optional[str] = None,
            link: Optional[str] = None,
            image: Optional[str] = None,
            prompt: Optional[str] = None,
        ):
            """
            构建一个ark37消息对象，用于发送消息

            :param title: 标题文本（选填）
            :param content: 内容文本（选填）
            :param link: 跳转的链接url（选填）
            :param image: 略缩图url，不可发送本地图片（选填）
            :param prompt: 消息弹窗通知的文本内容（选填）
            """
            super().__init__()
            self.__title = title
            self.__content = content
            self.__link = link
            self.__image = image
            self.__prompt = prompt

            self._constructed_obj = self.__construct(self._message_id, self._event_id)

        def __repr__(self):
            return (
                f"<MessageArk37 title={self.__title}, content={self.__content}, link={self.__link}, "
                f"image={self.__image}, prompt={self.__prompt}>"
            )

        def __construct(self, message_id, event_id) -> _MessageConstructRet:
            json_ = {
                "ark": {
                    "template_id": 37,
                    "kv": [
                        {"key": "#PROMPT#", "value": self.__prompt},
                        {"key": "#METATITLE#", "value": self.__title},
                        {"key": "#METASUBTITLE#", "value": self.__content},
                        {"key": "#METACOVER#", "value": self.__image},
                        {"key": "#METAURL#", "value": self.__link},
                    ],
                },
                "msg_id": message_id,
                "event_id": event_id,
            }
            return _MessageConstructRet(result=True, kwargs={"json": json_})

    class MessageMarkdown(BaseMessageApiModel):
        def __init__(
            self,
            template_id: Optional[str] = None,
            key_values: Optional[
                Union[
                    List[Dict[str, Union[str, List[str]]]],
                    Dict[str, Union[str, List[str]]],
                ]
            ] = None,
            content: Optional[str] = None,
            keyboard_id: Optional[str] = None,
            keyboard_content: Optional[Dict] = None,
        ):
            """
            构建一个markdown消息对象，用于发送消息

            :param template_id: markdown 模板 id（选填，与content不可同时存在）
            :param key_values: markdown 模版 key values列表，格式为：{key1: value1, key2: value2}（选填，与content不可同时存在）
            :param content: 原生 markdown 内容（选填，与template_id, key, values不可同时存在）
            :param keyboard_id: keyboard 模板 id（选填，与keyboard_content不可同时存在）
            :param keyboard_content: 原生 keyboard 内容（选填，与keyboard_id不可同时存在）
            """
            super().__init__()
            self.__template_id = template_id
            self.__key_values = key_values
            self.__content = content
            self.__keyboard_id = keyboard_id
            self.__keyboard_content = keyboard_content

            self._constructed_obj = self.__construct(self._message_id, self._event_id)

        def __repr__(self):
            return (
                f"<MessageMarkdown template_id={self.__template_id}, key_values={self.__key_values}, "
                f"content={self.__content}, keyboard_id={self.__keyboard_id}, "
                f"keyboard_content={self.__keyboard_content}>"
            )

        def __construct(self, message_id, event_id) -> _MessageConstructRet:
            """
            internal construct method

            :return: kwargs for http api request
            """
            logger_msg = None
            if self.__keyboard_content:
                if self.__keyboard_id:
                    logger_msg = "注意keyboard_id与keyboard_content不可同时存在，注意系统已根据优先级仅保留keyboard_content"
                keyboard = {"content": self.__keyboard_content}
            elif self.__keyboard_id:
                keyboard = {"id": self.__keyboard_id}
            else:
                keyboard = None
            if self.__content:
                if self.__template_id:
                    logger_msg = "注意content与template_id不可同时存在，注意系统已根据优先级仅保留content"
                json_ = {
                    "markdown": {"content": self.__content},
                    "msg_id": message_id,
                    "event_id": event_id,
                    "keyboard": keyboard,
                }
            else:
                if not self.__template_id or not self.__key_values:
                    return _MessageConstructRet(
                        result=False,
                        error_ret=sdk_error_temp(
                            "注意content与template_id必须存在任意一个，否则消息无法下发！"
                        ),
                    )
                if isinstance(self.__key_values, dict):
                    params = []
                    for k, v in self.__key_values.items():
                        if isinstance(v, list) or isinstance(v, tuple):
                            v = list(v)
                        else:
                            v = [str(v)]
                        params.append({"key": k, "values": v})
                else:
                    try:
                        params = list(self.__key_values)
                        for items in params:
                            for k, v in items.items():
                                if isinstance(v, list) or isinstance(v, tuple):
                                    items[k] = [str(v[0])]
                                else:
                                    items[k] = [str(v)]
                    except Exception:
                        return _MessageConstructRet(
                            result=False,
                            error_ret=sdk_error_temp(
                                "发送markdown消息的key_values仅可以是dict或list[dict]类型！"
                            ),
                        )
                json_ = {
                    "markdown": {
                        "custom_template_id": self.__template_id,
                        "params": params,
                    },
                    "msg_id": message_id,
                    "event_id": event_id,
                    "keyboard": keyboard,
                }
            return _MessageConstructRet(
                result=True, logger_msg=logger_msg, kwargs={"json": json_}
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pytest

import qg_botsdk

from ._base_context import bg_bot, config


def data_assertion(data):
    assert getattr(data, "result", None)
    assert hasattr(data, "trace_id")
    assert hasattr(data, "data")


@pytest.mark.run_order(3)
class TestAPI:
    _params = {}

    @pytest.mark.timeout(10)
    def test_security_check(self, bg_bot):
        assert bg_bot.api.security_check("abc")
        assert not bg_bot.api.security_check("操你妈")

    @pytest.mark.timeout(10)
    def test_get_bot_id(self, bg_bot):
        with pytest.raises(DeprecationWarning):
            bg_bot.api.get_bot_id()

    @pytest.mark.timeout(10)
    def test_get_bot_info(self, bg_bot):
        ret = bg_bot.api.get_bot_info()
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_get_bot_guilds(self, bg_bot):
        ret = bg_bot.api.get_bot_guilds()
        data_assertion(ret)
        self._params["guild_id"] = ret.data[0].id

    @pytest.mark.timeout(10)
    def test_get_guild_info(self, bg_bot):
        ret = bg_bot.api.get_guild_info(self._params.get("guild_id"))
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_get_guild_channels(self, bg_bot):
        ret = bg_bot.api.get_guild_channels(self._params.get("guild_id"))
        data_assertion(ret)
        for item in ret.data:
            if item.type == 0:
                self._params["channel_id"] = item.id

    @pytest.mark.timeout(10)
    def test_get_channels_info(self, bg_bot):
        ret = bg_bot.api.get_channels_info(self._params.get("channel_id"))
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_create_channels(self, bg_bot):
        ret = bg_bot.api.create_channels(
            self._params.get("guild_id"),
            "test0",
            0,
            0,
        )
        data_assertion(ret)
        self._params["channel_id"] = ret.data.id

    # skipped: patch_channels

    @pytest.mark.timeout(10)
    def test_delete_channels(self, bg_bot):
        ret = bg_bot.api.delete_channels(self._params.get("channel_id"))
        data_assertion(ret)

    # skipped: get_online_nums

    @pytest.mark.timeout(10)
    def test_get_guild_members(self, bg_bot):
        ret = bg_bot.api.get_guild_members(self._params.get("guild_id"))
        data_assertion(ret)
        self._params["user_id"] = ret.data[0].user.id

    @pytest.mark.timeout(10)
    def test_get_role_members(self, bg_bot):
        ret = bg_bot.api.get_role_members(self._params.get("guild_id"), "2")
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_get_member_info(self, bg_bot):
        ret = bg_bot.api.get_member_info(
            self._params.get("guild_id"), self._params.get("user_id")
        )
        data_assertion(ret)

    # skipped: delete_member

    @pytest.mark.timeout(10)
    def test_get_guild_roles(self, bg_bot):
        ret = bg_bot.api.get_guild_roles(self._params.get("guild_id"))
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_create_role(self, bg_bot):
        ret = bg_bot.api.create_role(self._params.get("guild_id"), "test0")
        data_assertion(ret)
        self._params["role_id"] = ret.data.role_id

    @pytest.mark.timeout(10)
    def test_patch_role(self, bg_bot):
        ret = bg_bot.api.patch_role(
            self._params.get("guild_id"), self._params.get("role_id"), "test1"
        )
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_delete_role(self, bg_bot):
        ret = bg_bot.api.delete_role(
            self._params.get("guild_id"), self._params.get("role_id")
        )
        data_assertion(ret)

    # skipped: create_role_member, delete_role_member

    # skipped: get_channel_member_permission, get_channel_role_permission

    # skipped: put_channel_member_permission, put_channel_role_permission

    @pytest.mark.timeout(10)
    def test_send_msg(self, bg_bot):
        ret = bg_bot.api.send_msg(self._params.get("channel_id"), "test")
        data_assertion(ret)
        self._params["message_id"] = ret.data.id

    @pytest.mark.timeout(10)
    def test_get_message_info(self, bg_bot):
        ret = bg_bot.api.get_message_info(
            self._params.get("channel_id"), self._params.get("message_id")
        )
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_send_embed(self, bg_bot):
        ret = bg_bot.api.send_embed(self._params.get("channel_id"), "test", "test")
        data_assertion(ret)
        self._params["message_id"] = ret.data.id

    # skipped: send_ark_23, send_ark_24, send_ark_37, send_markdown

    @pytest.mark.timeout(10)
    def test_get_guild_setting(self, bg_bot):
        ret = bg_bot.api.get_guild_setting(self._params.get("guild_id"))
        data_assertion(ret)

    # skipped: create_dm_guild, send_dm, delete_dm_msg
    # skipped: mute_all_member, mute_member, mute_members

    @pytest.mark.timeout(10)
    def test_create_announce(self, bg_bot):
        ret = bg_bot.api.create_announce(
            self._params.get("guild_id"),
            "test",
            recommend_channels_id=[self._params.get("channel_id")],
            recommend_channels_introduce=["testing"],
        )
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_delete_announce(self, bg_bot):
        ret = bg_bot.api.delete_announce(self._params.get("guild_id"))
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_create_pinmsg(self, bg_bot):
        ret = bg_bot.api.create_pinmsg(
            self._params.get("channel_id"), self._params.get("message_id")
        )
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_delete_pinmsg(self, bg_bot):
        ret = bg_bot.api.delete_pinmsg(
            self._params.get("channel_id"), self._params.get("message_id")
        )
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_get_pinmsg(self, bg_bot):
        ret = bg_bot.api.get_pinmsg(self._params.get("channel_id"))
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_get_schedules(self, bg_bot):
        ret = bg_bot.api.get_schedules(self._params.get("guild_id"))
        data_assertion(ret)

    # skipped: test_create_schedule, create_schedule, patch_schedule, delete_schedule

    @pytest.mark.timeout(10)
    def test_create_reaction(self, bg_bot):
        ret = bg_bot.api.create_reaction(
            self._params.get("channel_id"),
            self._params.get("message_id"),
            "2",
            "128516",
        )
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_get_reaction_users(self, bg_bot):
        ret = bg_bot.api.get_reaction_users(
            self._params.get("channel_id"),
            self._params.get("message_id"),
            "2",
            "128516",
        )
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_delete_reaction(self, bg_bot):
        ret = bg_bot.api.delete_reaction(
            self._params.get("channel_id"),
            self._params.get("message_id"),
            "2",
            "128516",
        )
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_delete_msg(self, bg_bot):
        ret = bg_bot.api.delete_msg(
            self._params.get("channel_id"), self._params.get("message_id")
        )
        data_assertion(ret)

    # skipped: control_audio, bot_on_mic, bot_off_mic, get_threads, get_thread_info, create_thread, delete_thread

    @pytest.mark.timeout(10)
    def test_get_guild_permissions(self, bg_bot):
        ret = bg_bot.api.get_guild_permissions(self._params.get("guild_id"))
        data_assertion(ret)

    @pytest.mark.timeout(10)
    def test_create_permission_demand(self, bg_bot):
        ret = bg_bot.api.create_permission_demand(
            self._params.get("guild_id"), self._params.get("channel_id"), "create_role"
        )
        data_assertion(ret)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Dict, Hashable, Optional

from ._statics import TraceNames
from ._utils import TraceCallerData, stack_exception_handler
from .model import Scope, SessionObject, SessionStatus


class AbstractSessionManager:
    """
    Session Manager用于管理在不同作用域（scope）下的 sessions，主要用处为存储运行时的临时数据。

    注意：为允许session可以保存class实例等py数据类型，session的数据以pickle方式保存
    """

    # -*- session manage methods -*-
    def new(
        self,
        scope: Scope,
        key: Hashable,
        data: Dict = None,
        identify: Hashable = None,
        is_replace: bool = True,
        timeout: Optional[float] = None,
        timeout_reply: str = None,
        inactive_gc_timeout: Optional[float] = 0,
    ) -> SessionObject:
        """
        根据scope作用域和相应的key，创建一个新的session（可用于间保存和传递字典的k-v pair数据）

        :param scope: 作用域(USER, GUILD, CHANNEL, GLOBAL)
        :param key: session的键，要求为可哈希对象，根据键在作用域下创建session
        :param data: session保存的字典类型数据，可选
        :param identify: scope作用域下的标识，不填时从栈查找本次的消息数据并自动填入（如Scope.USER会为此项自动填入user id），可选
        :param is_replace: 是否替换已存在的session，False且已存在相应key的session时报错，可选
        :param timeout: session处于ACTIVE状态下转为INACTIVE的超时时间，单位秒，None为没有超时，可选
        :param timeout_reply: 当session超时（从ACTIVE状态转为INACTIVE时）的回复消息内容，可选
        :param inactive_gc_timeout: session处于INACTIVE状态后自动回收的时间，单位秒，None为没有超时，可选

        :return: 存储session的数据的SessionObject
        """
        ...

    def get(
        self,
        scope: Scope,
        key: Hashable,
        identify: Hashable = None,
        default=None,
    ) -> SessionObject:
        """
        根据scope作用域和key，查询并获取相应的session

        :param scope: 作用域(USER, GUILD, CHANNEL, GLOBAL)
        :param key: session的键，要求为可哈希对象，根据键在作用域下查找session
        :param identify: scope作用域下的标识，不填时从栈查找本次的消息数据并自动填入（如Scope.USER会为此项自动填入user id），可选
        :param default: 未找到时返回的默认值数据，可选

        :return: 存储session的数据的SessionObject
        """
        ...

    def update(
        self,
        scope: Scope,
        key: Hashable,
        data: Dict,
        identify: Hashable = None,
    ) -> SessionObject:
        """
        根据scope作用域和key，更新相应session的存储数据，具体表现为dict.update()

        :param scope: 作用域(USER, GUILD, CHANNEL, GLOBAL)
        :param key: session的键，要求为可哈希对象，根据键在作用域下查找session
        :param data: 需要更新进入session的典类型数据
        :param identify: scope作用域下的标识，不填时从栈查找本次的消息数据并自动填入（如Scope.USER会为此项自动填入user id），可选

        :return: 存储session的数据的SessionObject
        """
        ...

    def remove(
        self, scope: Scope = None, identify: Hashable = None, key: Hashable = None
    ):
        """
        根据scope作用域和key，清空相应session的数据

        :param scope: 作用域(USER, GUILD, CHANNEL, GLOBAL)，如不存在则清空所有session
        :param identify: scope作用域下的标识，要求为可哈希对象，如不存在则清空相应scope下的所有session
        :param key: session的键，要求为可哈希对象，根据键在作用域下查找session，如不存在则清空相应identify下的所有session

        :return: 存储session的数据的SessionObject
        """
        ...

    def end(
        self, scope: Scope, key, inactive_gc_timeout: Optional[float] = 0
    ) -> SessionObject:
        """
        根据scope作用域和key，结束相应session，并在inactive_gc_timeout后进行回收

        :param scope: 作用域(USER, GUILD, CHANNEL, GLOBAL)
        :param key: session的键，要求为可哈希对象，根据键在作用域下查找session
        :param inactive_gc_timeout: session处于INACTIVE状态后自动回收的时间，单位秒，可选
        """
        ...

    def get_all(self) -> Dict:
        """
        获取所有session
        """
        ...

    def set_status(
        self,
        scope: Scope,
        key: Hashable,
        status: SessionStatus,
        identify: Hashable = None,
    ) -> None:
        """
        根据scope作用域和key，设置相应session的状态

        :param scope: 作用域(USER, GUILD, CHANNEL, GLOBAL)
        :param key: session的键，要求为可哈希对象，根据键在作用域下查找session
        :param status: session的状态，类型为SessionModel.Status
        :param identify: scope作用域下的标识，不填时从栈查找本次的消息数据并自动填入（如Scope.USER会为此项自动填入user id），可选
        """
        ...

    def get_status(
        self, scope: Scope, key: Hashable, identify: Hashable = None
    ) -> SessionStatus:
        """
        根据scope作用域和key，获取相应session的状态

        :param scope: 作用域(USER, GUILD, CHANNEL, GLOBAL)
        :param key: session的键，要求为可哈希对象，根据键在作用域下查找session
        :param identify: scope作用域下的标识，不填时从栈查找本次的消息数据并自动填入（如Scope.USER会为此项自动填入user id），可选

        :return: session的状态，类型为SessionModel.Status
        """
        ...

    # -*- fetch/commit data methods -*-
    def fetch_data(self, is_info: bool = True):
        """
        从数据库（默认为session_data文件夹）中获取session数据，更新相应session的存储数据，具体表现为dict.update()

        :param is_info: 读取成功时是否log信息，默认为True
        """
        ...

    def commit_data(self, is_info: bool = True):
        """
        将session数据保存到数据库（默认为session_data文件夹）

        :param is_info: 读取成功时是否log信息，默认为True
        """
        ...

    def set_auto_commit(self, is_auto_commit: bool):
        """
        设置是否自动保存session数据到数据库（默认为session_data文件夹）

        :param is_auto_commit: 是否自动保存session数据到数据库
        """
        ...

    def set_commit_path(self, commit_path: str):
        """
        设置保存session数据的数据库路径

        :param commit_path: 保存session数据的数据库路径
        """
        ...


class SessionPatcher(AbstractSessionManager):
    """
    Patcher for session manager, currying data from stack and pass to actual function on session manager.
    """

    def __getattribute__(self, item):
        _locals = TraceCallerData(TraceNames, ()).get_target_data()
        _session = _locals["self"].session_manager
        func = getattr(_session, item)
        if not func:
            raise AttributeError(f"Plugins.session has no attribute '{item}'")

        if item in (
            "api",
            "start",
            "register_wait_for",
            "check_wait_for",
            "del_wait_for",
            "wait_for_message_checker",
        ):
            raise AttributeError(f"Plugins.session has no attribute '{item}'")
        elif item in (
            "remove",
            "fetch_data",
            "commit_data",
            "set_auto_commit",
            "set_commit_path",
        ):
            return func

        data_obj = _locals["args"][0]

        def wrap(*args, **kwargs):
            try:
                ret = func(data_obj, *args, **kwargs)
            except Exception as e:
                _handler = stack_exception_handler(e, 2)
                _logger = _locals["self"].logger
                _logger.error(e.__repr__())
                _logger.error(_handler)
                return None
            return ret

        return wrap



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
from asyncio import sleep

import pytest

from qg_botsdk._session import SessionManager
from qg_botsdk._utils import objectize
from qg_botsdk.logger import Logger
from qg_botsdk.model import Scope, SessionStatus

from ._base_context import bot_async, config

MockObj = objectize(
    {
        "author": {"avatar": "xxx", "bot": False, "id": "111", "username": "xxx"},
        "channel_id": "222",
        "content": "plugins_test",
        "treated_msg": "plugins_test",
        "guild_id": "333",
        "id": "444",
        "member": {"joined_at": "xxx", "nick": "xxx", "roles": []},
        "mentions": [],
        "seq": 9555,
        "seq_in_channel": "9555",
        "timestamp": "xxx",
        "t": "MESSAGE_CREATE",
    }
)


@pytest.fixture()
def session(bot_async):
    _session = SessionManager(Logger(config["bot_id"]))
    _session.api = bot_async.api
    _session.start(bot_async.loop)
    return _session


@pytest.mark.run_order(1)
class TestSession:
    @pytest.mark.timeout(10)
    def test_session_base(self, session):
        session.remove()
        _all_data = session.get_all()
        for _scope in _all_data:
            assert _all_data[_scope] == {}
        assert session.get(MockObj, Scope.USER, "test") is None
        session.new(MockObj, Scope.USER, "test", data={"test": "test"})
        session.new(MockObj, Scope.USER, "test", data={"test": "test"})
        session.new(MockObj, Scope.GUILD, "test", data={"test": "test"})
        with pytest.raises(KeyError):
            session.new(
                MockObj, Scope.USER, "test", data={"test": "test"}, is_replace=False
            )
        session.new(MockObj, Scope.USER, "test1", data={"test1": "test1"})
        session.new(MockObj, Scope.CHANNEL, "test", data={"test": "test"})
        session.new(MockObj, Scope.GLOBAL, "test", data={"test": "test"})
        _all_data = session.get_all()
        assert len(_all_data[Scope.USER.value]["111"]) == 2
        assert len(_all_data[Scope.GUILD.value]["333"]) == 1
        assert len(_all_data[Scope.CHANNEL.value]["222"]) == 1
        assert len(_all_data[Scope.GLOBAL.value]) == 1
        assert _all_data[Scope.USER.value]["111"]["test"].data == {"test": "test"}
        assert _all_data[Scope.USER.value]["111"]["test1"].data == {"test1": "test1"}
        assert _all_data[Scope.GUILD.value]["333"]["test"].data == {"test": "test"}
        assert _all_data[Scope.CHANNEL.value]["222"]["test"].data == {"test": "test"}
        assert _all_data[Scope.GLOBAL.value]["test"].data == {"test": "test"}
        session.remove(Scope.GUILD, "333", "test")
        assert session.get(MockObj, Scope.GUILD, "test") is None
        _all_data = session.get_all()
        assert len(_all_data[Scope.USER.value]["111"]) == 2
        assert len(_all_data[Scope.GUILD.value]["333"]) == 0
        assert len(_all_data[Scope.CHANNEL.value]["222"]) == 1
        assert len(_all_data[Scope.GLOBAL.value]) == 1
        session.remove(Scope.USER)
        print(_all_data[Scope.USER.value])
        assert len(session.get_all()[Scope.USER.value]) == 0
        session.remove()
        _all_data = session.get_all()
        for _scope in _all_data:
            assert _all_data[_scope] == {}

    @pytest.mark.timeout(10)
    def test_internal(self, session):
        with pytest.raises(ValueError):
            session._SessionManager__check_path("test.db")
        with pytest.raises(ValueError):
            session._SessionManager__check_path("__init__.py")
        assert session._SessionManager__valid_scope(Scope.USER) == "USER"
        assert session._SessionManager__valid_scope(Scope.USER.value) == "USER"
        with pytest.raises(ValueError):
            session._SessionManager__valid_scope("abc")

    @pytest.mark.timeout(10)
    def test_get_data(self, session):
        session_data_from_new = session.new(
            MockObj, Scope.USER, "test", data={"test": "test"}
        )
        session_data_from_get = session.get(MockObj, Scope.USER, "test")
        assert session_data_from_new.data == session_data_from_get.data
        assert session_data_from_new.scope == session_data_from_get.scope
        assert session_data_from_new.key == session_data_from_get.key
        assert session_data_from_new.status == session_data_from_get.status
        assert session_data_from_new.identify == session_data_from_get.identify
        assert session_data_from_get.scope == Scope.USER.value
        assert session_data_from_get.key == "test"
        assert session_data_from_get.status == SessionStatus.ACTIVE
        assert session_data_from_get.data == {"test": "test"}

    @pytest.mark.timeout(15)
    def test_timeout(self, bot_async, session):
        session.remove()
        session.new(MockObj, Scope.USER, "test", data={"test": "test"})
        assert (
            session.get_all()[Scope.USER.value]["111"]["test"].status
            == SessionStatus.ACTIVE
        )
        session.end(MockObj, Scope.USER, "test")
        assert (
            session.get_all()[Scope.USER.value]["111"]["test"].status
            == SessionStatus.INACTIVE
        )
        bot_async.loop.run_until_complete(sleep(0.5))
        assert "111" not in session.get_all()[Scope.USER.value]
        session.new(MockObj, Scope.USER, "test", data={"test": "test"}, timeout=0.1)
        assert (
            session.get_all()[Scope.USER.value]["111"]["test"].status
            == SessionStatus.ACTIVE
        )
        bot_async.loop.run_until_complete(sleep(0.5))
        assert "111" not in session.get_all()[Scope.USER.value]
        session.new(MockObj, Scope.USER, "test", data={"test": "test"})
        session.end(MockObj, Scope.USER, "test", inactive_gc_timeout=0.5)
        assert (
            session.get_all()[Scope.USER.value]["111"]["test"].status
            == SessionStatus.INACTIVE
        )
        bot_async.loop.run_until_complete(sleep(0.2))
        assert (
            session.get_all()[Scope.USER.value]["111"]["test"].status
            == SessionStatus.INACTIVE
        )
        bot_async.loop.run_until_complete(sleep(1))
        assert "111" not in session.get_all()[Scope.USER.value]

    @pytest.mark.timeout(10)
    def test_set_get_status(self, bot_async, session):
        session.remove()
        session.new(MockObj, Scope.USER, "test", data={"test": "test"})
        session.set_status(MockObj, Scope.USER, "test", SessionStatus.HANGING)
        assert session.get_status(MockObj, Scope.USER, "test") == SessionStatus.HANGING
        assert (
            session.get_all()[Scope.USER.value]["111"]["test"].status
            == SessionStatus.HANGING
        )
        assert session.get(MockObj, Scope.USER, "test").status == SessionStatus.ACTIVE
        assert session.get(MockObj, Scope.USER, "test").status == SessionStatus.ACTIVE
        session.set_status(MockObj, Scope.USER, "test", SessionStatus.INACTIVE)
        assert (
            session.get_all()[Scope.USER.value]["111"]["test"].status
            == SessionStatus.INACTIVE
        )
        bot_async.loop.run_until_complete(sleep(0.5))
        assert session.get(MockObj, Scope.USER, "test") is None

    @pytest.mark.timeout(10)
    def test_set_auto_commit(self, session):
        session.remove()
        assert session._SessionManager__is_auto_commit
        session.set_auto_commit(False)
        assert not session._SessionManager__is_auto_commit

    @pytest.mark.timeout(10)
    def test_commit(self, session):
        session.remove()
        _data = session.get_all()
        session.commit_data()
        session.fetch_data()
        assert session.get_all() == _data
        default_path = os.path.join(os.getcwd(), "session_data")
        assert os.path.exists(
            os.path.join(default_path, f"session_{config['bot_id']}.db")
        )
        new_path = os.path.join(default_path, "session_data")
        session.set_commit_path(new_path)
        session.commit_data()
        assert os.path.exists(os.path.join(new_path, f"session_{config['bot_id']}.db"))



class IdTokenMissing(Exception):
    """Raised when id or token is missing"""

    ...


class IdTokenError(Exception):
    """Raised when id or token is invalid"""

    ...


class WaitTimeoutError(Exception):
    """Raised when api.wait_for() is timeout"""

    ...


class WaitError(Exception):
    """Raised when api.wait_for() got an error"""

    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from io import StringIO
from os import getcwd, getenv
from threading import Thread
from typing import Dict

import pytest

import qg_botsdk

if getcwd().endswith("test"):
    read_path = "./config/config.yaml"
    plugin_path = "./_testing_plugin.py"
else:
    read_path = "./test/config/config.yaml"
    plugin_path = "./test/_testing_plugin.py"

try:
    config: Dict
    raw_config = getenv("TEST_CONFIG", None)
    if raw_config:
        config = qg_botsdk.utils.read_yaml(StringIO(raw_config))
    else:
        config = qg_botsdk.utils.read_yaml(read_path)
except FileNotFoundError:
    raise FileNotFoundError(
        "请先参考test/config/config.example.yaml，创建test/config/config.yaml文件进行单元测试"
    )


@pytest.fixture()
def bot():
    return qg_botsdk.BOT(
        bot_id=config["bot_id"],
        bot_token=config["bot_token"],
        is_private=True,
        is_sandbox=True,
    )


@pytest.fixture()
def bot_async():
    return qg_botsdk.BOT(
        bot_id=config["bot_id"],
        bot_token=config["bot_token"],
        is_private=True,
        is_async=True,
        is_sandbox=True,
    )


@pytest.fixture(scope="class")
def bg_bot() -> qg_botsdk.BOT:
    bot = qg_botsdk.BOT(
        bot_id=config["bot_id"],
        bot_token=config["bot_token"],
        is_sandbox=True,
        api_timeout=3,
    )
    Thread(target=bot.start, daemon=True).start()
    bot.security_setup(config["mini_id"], config["mini_token"])
    return bot


def _callback(data):
    data.reply("test")


async def _async_callback(data):
    await data.reply("test")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import asyncio
import os
from logging import getLogger
from re import compile as re_compile
from typing import List

import pytest

import qg_botsdk

from ._base_context import _callback, bot, bot_async, config, plugin_path


def _repeat_or_start():
    pass


async def _async_repeat_or_start():
    pass


async def _queue_task(start_time: List):
    start_time.append(asyncio.get_event_loop().time())
    await asyncio.sleep(0.5)


@pytest.mark.run_order(1)
class TestBase:
    @staticmethod
    @pytest.mark.timeout(5)
    def test_base():
        bot = qg_botsdk.BOT(bot_id=config["bot_id"], bot_token=config["bot_token"])
        assert bot.bot_id == config["bot_id"]
        assert bot.bot_token == config["bot_token"]
        assert not bot.is_async
        assert not bot.is_private
        assert bot.bot_url == r"https://api.sgroup.qq.com"
        assert bot.no_permission_warning
        assert bot._http_session._is_retry
        assert bot._http_session._is_log_error
        assert bot._shard_no == 0
        assert bot._total_shard == 1
        assert bot.max_workers == 32
        assert bot._http_session._queue._MAX_RUNNING_SLOTS == 0
        assert bot._http_session._timeout.total == 20
        assert bot._intents == 0
        assert bot.bot_headers == {
            "Authorization": f'Bot {config["bot_id"]}.{config["bot_token"]}'
        }
        assert (
            str(bot)
            == f'<qg_botsdk.BOT object [id: {config["bot_id"]}, token: {config["bot_token"]}]>'
        )

        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"], bot_token=config["bot_token"], is_async=True
        )
        assert bot.is_async
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"], bot_token=config["bot_token"], is_private=True
        )
        assert bot.is_private
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"], bot_token=config["bot_token"], is_sandbox=True
        )
        assert bot.bot_url == r"https://sandbox.api.sgroup.qq.com"
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"],
            bot_token=config["bot_token"],
            no_permission_warning=False,
        )
        assert not bot.no_permission_warning
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"], bot_token=config["bot_token"], is_retry=False
        )
        assert not bot._http_session._is_retry
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"], bot_token=config["bot_token"], is_log_error=False
        )
        assert not bot._http_session._is_log_error
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"],
            bot_token=config["bot_token"],
            shard_no=0,
            total_shard=2,
        )
        assert bot._shard_no == 0
        assert bot._total_shard == 2
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"],
            bot_token=config["bot_token"],
            shard_no=1,
            total_shard=2,
        )
        assert bot._shard_no == 1
        assert bot._total_shard == 2
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"], bot_token=config["bot_token"], max_workers=64
        )
        assert bot.max_workers == 64
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"],
            bot_token=config["bot_token"],
            api_max_concurrency=5,
        )
        assert bot._http_session._queue._MAX_RUNNING_SLOTS == 5
        bot = qg_botsdk.BOT(
            bot_id=config["bot_id"], bot_token=config["bot_token"], api_timeout=60
        )
        assert bot._http_session._timeout.total == 60

        with pytest.raises(DeprecationWarning):
            bot = qg_botsdk.BOT(
                bot_id=config["bot_id"], bot_token=config["bot_token"], bot_secret="xxx"
            )
        with pytest.raises(Exception) as e:
            bot = qg_botsdk.BOT(bot_id="", bot_token="")
            assert e.__name__ == "IdTokenMissing"
        with pytest.raises(Exception) as e:
            bot = qg_botsdk.BOT(bot_id="xxx", bot_token="xxx")
            bot.start()
            assert e.__name__ == "IdTokenError"

        assert "__sdk_default_logger" not in qg_botsdk.Plugins.get_preprocessor_names()
        bot.load_default_msg_logger()
        preprocessors = [x for x in qg_botsdk.Plugins.get_preprocessor_names()]
        assert "__sdk_default_logger" in preprocessors
        bot.load_default_msg_logger()
        assert preprocessors == [x for x in qg_botsdk.Plugins.get_preprocessor_names()]

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_msg(bot):
        bot.is_private = False
        bot.bind_msg(_callback)
        assert bot._intents == 1 << 30
        bot._intents = 0
        assert bot._func_registers.get("on_msg", None) == _callback
        assert bot.msg_treat

        bot.bind_msg(_callback, all_msg=True)
        assert bot._intents == 1 << 9
        bot._intents = 0

        bot.bind_msg(_callback, all_msg=False)
        assert bot._intents == 1 << 30
        bot._intents = 0

        bot.is_private = True
        bot.bind_msg(_callback, treated_data=False)
        assert bot._intents == 1 << 9
        assert not bot.msg_treat

        bot._func_registers = {}
        bot.bind_msg()(_callback)  # test decorator
        assert bot._func_registers.get("on_msg", None) == _callback

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_dm(bot):
        bot.bind_dm(_callback)
        assert bot._intents == 1 << 12
        assert bot.dm_treat
        assert bot._func_registers.get("on_dm", None) == _callback
        bot._intents = 0

        bot.bind_dm(_callback, treated_data=False)
        assert not bot.dm_treat

        bot._func_registers = {}
        bot.bind_dm()(_callback)  # test decorator
        assert bot._func_registers.get("on_dm", None) == _callback

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_msg_delete(bot):
        bot.is_private = False
        bot.bind_msg_delete(_callback)
        assert bot._intents == 1 << 30
        bot._intents = 0

        bot.is_private = True
        bot.bind_msg_delete(_callback)
        assert bot._intents == 1 << 9

        del_is_filter_self = bot._func_registers.get("del_is_filter_self", None)
        assert del_is_filter_self is not None and del_is_filter_self
        assert bot._func_registers.get("on_delete", None) == _callback

        bot.bind_msg_delete(_callback, is_filter_self=False)
        del_is_filter_self = bot._func_registers.get("del_is_filter_self", None)
        assert del_is_filter_self is not None and not del_is_filter_self

        bot._func_registers = {}
        bot.bind_msg_delete()(_callback)  # test decorator
        assert bot._func_registers.get("on_delete", None) == _callback

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_guild_event(bot):
        bot.bind_guild_event(_callback)
        assert bot._intents == 1
        assert bot._func_registers.get("on_guild_event", None) == _callback

        bot._func_registers = {}
        bot.bind_guild_event()(_callback)  # test decorator
        assert bot._func_registers.get("on_guild_event", None) == _callback

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_channel_event(bot):
        bot.bind_channel_event(_callback)
        assert bot._intents == 1
        assert bot._func_registers.get("on_channel_event", None) == _callback

        bot._func_registers = {}
        bot.bind_channel_event()(_callback)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_guild_member(bot):
        bot.bind_guild_member(_callback)
        assert bot._intents == 1 << 1
        assert bot._func_registers.get("on_guild_member", None) == _callback

        bot._func_registers = {}
        bot.bind_guild_member()(_callback)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_reaction(bot):
        bot.bind_reaction(_callback)
        assert bot._intents == 1 << 10
        assert bot._func_registers.get("on_reaction", None) == _callback

        bot._func_registers = {}
        bot.bind_reaction()(_callback)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_interaction(bot):
        bot.bind_interaction(_callback)
        assert bot._intents == 1 << 26
        assert bot._func_registers.get("on_interaction", None) == _callback

        bot._func_registers = {}
        bot.bind_interaction()(_callback)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_audit(bot):
        bot.bind_audit(_callback)
        assert bot._intents == 1 << 27
        assert bot._func_registers.get("on_audit", None) == _callback

        bot._func_registers = {}
        bot.bind_audit()(_callback)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_forum(bot):
        bot.bind_forum(_callback)
        assert bot._intents == 1 << 28
        assert bot._func_registers.get("on_forum", None) == _callback

        bot._func_registers = {}
        bot.bind_forum()(_callback)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_open_forum(bot):
        bot.bind_open_forum(_callback)
        assert bot._intents == 1 << 18
        assert bot._func_registers.get("on_open_forum", None) == _callback

        bot._func_registers = {}
        bot.bind_open_forum()(_callback)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_audio(bot):
        bot.bind_audio(_callback)
        assert bot._intents == 1 << 29
        assert bot._func_registers.get("on_audio", None) == _callback

        bot._func_registers = {}
        bot.bind_audio()(_callback)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_bind_live_channel_member(bot):
        bot.bind_live_channel_member(_callback)
        assert bot._intents == 1 << 19
        assert bot._func_registers.get("on_live_channel_member", None) == _callback

        bot._func_registers = {}
        bot.bind_live_channel_member()(_callback)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_register_repeat_event(bot):
        bot.register_repeat_event(_repeat_or_start)
        assert bot._repeat_function == _repeat_or_start

        bot._repeat_function = None
        bot.register_repeat_event()(_repeat_or_start)
        assert bot._repeat_function == _repeat_or_start

    @staticmethod
    @pytest.mark.timeout(5)
    def test_register_start_event(bot):
        bot.register_start_event(_repeat_or_start)
        assert bot._on_start_function == _repeat_or_start

        bot._on_start_function = None
        bot.register_start_event()(_repeat_or_start)
        assert bot._on_start_function == _repeat_or_start

        with pytest.raises(TypeError):
            bot.register_start_event(_async_repeat_or_start)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_async_register_start_event(bot_async):
        bot_async.register_start_event(_async_repeat_or_start)
        assert bot_async._on_start_function == _async_repeat_or_start

        bot_async._on_start_function = None
        bot_async.register_start_event()(_async_repeat_or_start)
        assert bot_async._on_start_function == _async_repeat_or_start

        with pytest.raises(TypeError):
            bot_async.register_start_event(_repeat_or_start)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_security_setup(bot):
        bot.security_setup(config["mini_id"], config["mini_token"])

    @staticmethod
    def plugin_assert(bot):
        assert [x for x in qg_botsdk.Plugins.get_commands_names()] == ["plugins_test"]
        assert [x for x in qg_botsdk.Plugins.get_preprocessor_names()] == [
            "before_command_test"
        ]
        plugins_list = bot._retrieve_new_plugins()
        assert len(plugins_list) == 2
        command_list = plugins_list[0]
        before_command_list = plugins_list[1]
        assert len(command_list) == 1
        assert len(before_command_list) == 1
        assert command_list[0].command == ["plugins_test"]

    @pytest.mark.timeout(5)
    def test_load_plugins(self, bot):
        bot._retrieve_new_plugins()  # clear plugins before first

        from . import _testing_plugin

        self.plugin_assert(bot)

        assert bot._retrieve_new_plugins() == ([], [])

        bot.load_plugins(plugin_path)
        self.plugin_assert(bot)

        with pytest.raises(ModuleNotFoundError):
            bot.load_plugins("xxx")

    @staticmethod
    def direct_plugin_assert(bot, regex=False):
        assert [x for x in qg_botsdk.Plugins.get_commands_names()] == ["_callback"]
        assert [x for x in qg_botsdk.Plugins.get_preprocessor_names()] == ["<lambda>"]
        plugins_list = bot._retrieve_new_plugins()
        assert len(plugins_list) == 2
        command_list = plugins_list[0]
        before_command_list = plugins_list[1]
        assert len(command_list) == 1
        assert len(before_command_list) == 1
        if regex:
            assert command_list[0].regex == [re_compile(r"command_(\s)")]
        else:
            assert command_list[0].command == ["abc"]

    @pytest.mark.timeout(5)
    def test_direct_load_plugins(self, bot):
        bot._retrieve_new_plugins()  # clear plugins before first

        bot.on_command("abc")(_callback)
        bot.before_command()(lambda x: x)
        self.direct_plugin_assert(bot)

        assert bot._retrieve_new_plugins() == ([], [])

        bot.on_command(["abc"])(_callback)
        bot.before_command()(lambda x: x)
        self.direct_plugin_assert(bot)

        assert bot._retrieve_new_plugins() == ([], [])

        bot.on_command(regex=r"command_(\s)")(_callback)
        bot.before_command()(lambda x: x)
        self.direct_plugin_assert(bot, regex=True)

        assert bot._retrieve_new_plugins() == ([], [])

        bot.on_command(regex=re_compile(r"command_(\s)"))(_callback)
        bot.before_command()(lambda x: x)
        self.direct_plugin_assert(bot, regex=True)

    @staticmethod
    @pytest.mark.timeout(5)
    def test_logger(caplog):
        _log_path = "./log/test_log_file"
        if os.path.exists(_log_path):
            for _file in os.listdir(_log_path):
                os.remove(os.path.join(_log_path, _file))
            os.rmdir(_log_path)
        logger = qg_botsdk.Logger("test_logger", _log_path)
        logger.info("test")
        assert os.path.exists(_log_path)
        assert len(os.listdir(_log_path)) == 1

        logger = qg_botsdk.Logger("test_logger", _log_path)
        logger.set_formatter(info_format="[%(levelname)s] %(message)s")
        logger.info("test")
        assert os.path.exists(_log_path)
        assert len(os.listdir(_log_path)) == 1

        logger.disable_logger("a")
        assert getLogger("a").disabled
        logger.disable_logger(["b", "c"])
        assert getLogger("b").disabled
        assert getLogger("c").disabled

    @staticmethod
    @pytest.mark.timeout(5)
    def test_queue():
        queue = qg_botsdk._queue.Queue(1)
        loop = asyncio.get_event_loop()
        start_time_arr = []
        loop.create_task(queue.create_task(_queue_task, start_time_arr))
        loop.run_until_complete(asyncio.sleep(0.1))
        loop.run_until_complete(queue.create_task(_queue_task, start_time_arr))
        assert start_time_arr[1] - start_time_arr[0] >= 0.5

    @staticmethod
    @pytest.mark.timeout(5)
    def test_func_checking(bot):
        with pytest.raises(TypeError):
            bot.bind_msg()(lambda x, y, z: print(x, y, z))

    @staticmethod
    @pytest.mark.timeout(5)
    def test_convert_color_util():
        assert qg_botsdk.utils.convert_color((255, 254, 253)) == 16645887
        assert qg_botsdk.utils.convert_color("#fffefd") == 16645887
        assert qg_botsdk.utils.convert_color("#FFFEFD") == 16645887
        with pytest.raises(TypeError):
            qg_botsdk.utils.convert_color("fffefg")
        with pytest.raises(TypeError):
            qg_botsdk.utils.convert_color("fffefdd")
        with pytest.raises(TypeError):
            qg_botsdk.utils.convert_color((255, 254, 253, 252))
        with pytest.raises(TypeError):
            qg_botsdk.utils.convert_color((255, 254, 256))
        with pytest.raises(TypeError):
            qg_botsdk.utils.convert_color(255)
